;; Generated package description from cond-star.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "cond-star" "1.0.0.20260523.114724" "Extended form of `cond' construct" '((emacs "24.3")) :commit "a96fc7d5465f97908a9a6d5d2cf3be8cbf6b0ebe" :url "https://elpa.gnu.org/packages/cond-star.html" :maintainer '("Richard Stallman" . "rms@gnu.org"))
